// Fragment shader for per-pixel Phong interpolation and shading.

// The "varying" keyword means that the parameter's value is interpolated
// between the nearby vertices.
varying vec3 N;
varying vec3 v;

uniform float brightness;

void main()
{
    // The scene's ambient light.
    vec4 ambient = gl_LightModel.ambient * gl_FrontMaterial.ambient;

	// The normal vectors is generally not normalized after being
	// interpolated across a triangle.  Here we normalize it.
	vec3 Normal = normalize(N);

	// Since the vertex is in eye space, the direction to the
	// viewer is simply the normalized vector from v to the
	// origin.
	vec3 Viewer = -normalize(v);

	// Get the lighting direction and normalize it.
	vec3 Light_Dir  = normalize(gl_LightSource[0].position.xyz);
	vec3 Light_Pnt_Original = gl_LightSource[1].position.xyz - v;
	vec3 Light_Pnt  = normalize(Light_Pnt_Original);

	// Compute halfway vector
	vec3 Half_Dir = normalize(Viewer+Light_Dir);
	vec3 Half_Pnt = normalize(Viewer+Light_Pnt);

	// Compute factor to prevent light leakage from below the
	// surface
	float B_Dir = 1.0;
	if(dot(Normal, Light_Dir)<0.0) B_Dir = 0.0;

	float B_Pnt = 1.0;
	if(dot(Normal, Light_Pnt)<0.0) B_Pnt = 0.0;

	// Compute geometric terms of diffuse and specular
	float diffuseShade_Dir = B_Dir * max(dot(Normal, Light_Dir), 0.0);
	float specularShade_Dir = 
	  B_Dir * pow(max(dot(Half_Dir, Normal), 0.0), gl_FrontMaterial.shininess);

	float diffuseShade_Pnt = B_Pnt * max(dot(Normal, Light_Pnt), 0.0);
	float specularShade_Pnt = 
	  B_Pnt * pow(max(dot(Half_Pnt, Normal), 0.0), gl_FrontMaterial.shininess);

	// Get the attenuation
	float dist = length(Light_Pnt_Original);
	float a = 1.0 / (gl_LightSource[1].constantAttenuation +
                gl_LightSource[1].linearAttenuation * dist +
                gl_LightSource[1].quadraticAttenuation * dist * dist);

	// Compute product of geometric terms with material and
	// lighting values
	vec4 diffuse = diffuseShade_Dir * gl_FrontLightProduct[0].diffuse;
	vec4 specular = specularShade_Dir * gl_FrontLightProduct[0].specular;
	ambient += gl_FrontLightProduct[0].ambient;

	diffuse += a * diffuseShade_Pnt * gl_FrontLightProduct[1].diffuse;
	specular += a * specularShade_Pnt * gl_FrontLightProduct[1].specular;
	ambient += a * gl_FrontLightProduct[1].ambient;

	// Assign final color
	gl_FragColor = ambient + diffuse + specular + gl_FrontMaterial.emission;
	// Scale with brightness value
	gl_FragColor = gl_FragColor * brightness;

}